import datetime
import os
import pickle

import numpy as np
import pandas as pd


class Session:
    def __init__(self, path, session_id, psg_id):
        """
        Class for holding all the data related to one session. Is currently dependent on PSG being part of the session

        :param str path: path to the processed data, should contain registry.pkl fil and the somnofy/ and psg/ folders.
        :param str session_id: ID of the session in question, something like 'VTNJQxIGChYWEgAA'.
        :param str psg_id: ID of the psg profile matching the session, something like '7412'.
        :param bool exclude_wake_before_sleep_onset: should epochs before sleep onset be removed, Default False.
        """

        # Initialize directly from input parameters
        self.path = path
        self.session_id = session_id
        self.psg_id = psg_id

        # track which data has been read
        self.loaded_data_frames = []
        # List of dataframes that have been loaded. Data can be reached with self.key_name

        # load data
        # SOMNOFY
        # read somnofy vitalsigns and environment
        self.somnofy = {}  # Dictionary containing both vitalsigns and environment

        self.vitalsigns = self.read_vitalsigns()
        self.loaded_data_frames.append('vitalsigns')

        self.somnofy = {}  # Clean memory

        #         self.respiration = self.read_respiration()
        #         self.loaded_data_frames.append('respiration')

        #         self.movement_instant = self.read_movement_instant()
        #         self.loaded_data_frames.append('movement_instant')

        self.derived = self.read_derived()
        self.loaded_data_frames.append('derived')

        # PSG
        # read time sync between psg and somnofy
        self.time_sync = self.read_time_sync()

        # read psg hypnograms
        self.psg = self.read_psg()

        # self.automatic_psg_hypnogram, self.automatic_psg_hypnogram_reliability = self.read_psg_auto()
        # self.loaded_data_frames += ['automatic_psg_hypnogram', 'automatic_psg_hypnogram_reliability']

        self.psg_analysed, self.psg_flow_events, self.spo2_flow_events = self.read_psg_analysed()
        self.loaded_data_frames.append('psg_analysed')
        self.loaded_data_frames.append('psg_flow_events')
        self.loaded_data_frames.append('spo2_flow_events')

        self.psg_derived = self.read_psg_derived()
        self.loaded_data_frames.append('psg_derived')

        self.psg_raw = self.read_psg_raw()
        self.loaded_data_frames.append('psg_raw')

        self.spo2 = self.read_spo2()
        self.loaded_data_frames.append('spo2')

        # Reindex dataframes to common index for psg and somnofy. Exclude wake time before sleep onset if True

        self.clip_data()

    def read_time_sync(self):
        """
        Read the time_sync pickle

        :return:
        :rtype: dict
        """
        time_sync_file_name = "somnofy_timesync_{}.pkl"
        time_sync = pickle.load(
            open(os.path.join(self.path, 'somnofy/', time_sync_file_name.format(self.session_id)), 'rb'))
        return time_sync

    def read_psg(self):
        """
        Reads the PSG pickle

        :return:
        :rtype: pd.DataFrame
        """
        psg_file_name = "hypnogram_{}.pkl"
        psg_data = pickle.load(open(self.path + '/psg/' + psg_file_name.format(self.psg_id), 'rb'))
        if not len(psg_data['sleep_analysis']['manual']) is 0:
            psg = pd.DataFrame(
                {k: psg_data['sleep_analysis']['manual'][0]['hypnogram'][k] for k in ['timestamp', 'sleep_stage']})
        elif 'automatic' in psg_data['hypnogram'].keys():
            psg = pd.DataFrame(
                {k: psg_data['sleep_analysis']['automatic'][0]['hypnogram'][k] for k in ['timestamp', 'sleep_stage']})
        else:
            raise ValueError('No hypnogram in the psg pickle')

        psg['timestamp'] = self.sync_psg(psg['timestamp'])
        return psg

    def read_vitalsigns(self):
        """
        Reads the vitalsigns portion of the somnofy_data pickle.

        :return:
        :rtype: pd.DataFrame
        """
        if not self.somnofy:
            self.somnofy = self.read_somnofy()
        vitalsigns = pd.DataFrame(self.somnofy['vitalsigns'])
        vitalsigns['respiration_rate'] = vitalsigns['respiration_rate'].shift(-1)  # Why shift -10?
        vitalsigns = vitalsigns.drop(vitalsigns.index[-10:])
        return vitalsigns

    def read_environment(self):
        """
        Reads the environment portion of the somnofy data pickle.

        :return:
        :rtype: pd.DataFrame
        """
        if not self.somnofy:
            self.somnofy = self.read_somnofy()
        environment = pd.DataFrame(self.somnofy['environment'])
        return environment

    def read_respiration(self):
        """
        Reads the respiration waveform pickle

        :return:
        :rtype: pd.DataFrame
        """
        respiration_file_name = "somnofy_data_{}_respirationwaveform.pkl"
        respiration_data = pickle.load(
            open(self.path + '/somnofy/' + respiration_file_name.format(self.session_id), 'rb'))
        respiration = pd.DataFrame(respiration_data['respirationwaveform'])
        return respiration

    def read_movement_instant(self):
        """
        Reads the movementinstant pickle.

        :return:
        :rtype: pd.DataFrame
        """
        movement_instant_file_name = "somnofy_data_{}_movementinstant.pkl"
        movement_instant_full = pickle.load(
            open(self.path + '/somnofy/' + movement_instant_file_name.format(self.session_id), 'rb'))
        movement_instant = pd.DataFrame(movement_instant_full['movement_instant'])
        return movement_instant

    def read_derived(self):
        """
        Reads the derived pickle.
        
        :return:
        :rtype: pd.DataFrame
        """
        derived_instant_file_name = "somnofy_data_{}_derived.pkl"
        derived_instant_full = pickle.load(
            open(self.path + '/somnofy/' + derived_instant_file_name.format(self.session_id), 'rb'))
        derived_instant = pd.DataFrame(derived_instant_full['derived']['respirationwaveform_normalized'])
        return derived_instant

    def read_somnofy(self):
        """
        Reads the somnofy_data pickle.

        :return:
        :rtype: dict
        """
        somnofy_file_name = "somnofy_data_{}.pkl"
        somnofy = pickle.load(
            open(os.path.join(self.path, 'somnofy/', somnofy_file_name.format(self.session_id)), 'rb'))
        return somnofy

    def read_psg_auto(self):
        """
        Reads the psg pickle.

        :return:
        :rtype: (pd.DataFrame, pd.DataFrame)
        """
        psg_file_name = "hypnogram_{}.pkl"
        psg_data = pickle.load(open(self.path + '/psg/' + psg_file_name.format(self.psg_id), 'rb'))
        automatic_psg_hypnogram = pd.DataFrame()

        try:
            automatic_psg_hypnogram['hypnogram'] = psg_data['sleep_analysis']['automatic'][0]['hypnogram'][
                'sleep_stage']
            automatic_psg_hypnogram['timestamp'] = psg_data['sleep_analysis']['automatic'][0]['hypnogram']['timestamp']
        except:
            pass

        automatic_psg_hypnogram_reliability = pd.DataFrame()
        automatic_psg_hypnogram_reliability['reliability'] = \
            psg_data['sleep_analysis']['automatic'][0]['hypnogram_reliability']['reliability']
        automatic_psg_hypnogram_reliability['timestamp'] = \
            psg_data['sleep_analysis']['automatic'][0]['hypnogram_reliability']['timestamp']
        automatic_psg_hypnogram['timestamp'] = self.sync_psg(automatic_psg_hypnogram['timestamp'])
        automatic_psg_hypnogram_reliability['timestamp'] = self.sync_psg(
            automatic_psg_hypnogram_reliability['timestamp'])
        return automatic_psg_hypnogram, automatic_psg_hypnogram_reliability

    def read_psg_analysed(self):
        """
        Reads the psg_analysed pickle.

        :return:
        :rtype: pd.DataFrame
        """
        psg_analysed_file_name = "psg_analysed_{}.pkl"
        psg_analysed_data = pickle.load(open(self.path + '/psg/' + psg_analysed_file_name.format(self.psg_id), 'rb'))
        psg_analysed = pd.DataFrame()

        psg_analysed['heart_rate'] = psg_analysed_data['analysed']['heart_rate']['data']
        psg_analysed['timestamp'] = psg_analysed_data['analysed']['heart_rate']['timestamp']
        psg_analysed['timestamp'] = self.sync_psg(psg_analysed['timestamp'])

        psg_flow_events = pd.DataFrame()
        psg_flow_events['data'] = psg_analysed_data['analysed']['flow_events']['data']
        psg_flow_events['category'] = psg_analysed_data['analysed']['flow_events']['category']

        # Represented in the data as a list of tuples with start/end time 
        time_stamps_flow = psg_analysed_data['analysed']['flow_events']['timestamp']

        # Getting start times for the events 
        timestamps = [time_range[0] for time_range in time_stamps_flow]
        psg_flow_events['timestamp'] = timestamps

        # Getting end times for the events 
        end_timestamps = [time_range[1] for time_range in time_stamps_flow]
        psg_flow_events['timestamp_end'] = end_timestamps

        # Syncing timestamps
        psg_flow_events['timestamp'] = self.sync_psg(psg_flow_events['timestamp'])
        psg_flow_events['timestamp_end'] = self.sync_psg(psg_flow_events['timestamp_end'])

        # SpO2 events
        spo2_flow_events = pd.DataFrame()

        spo2_flow_events['data'] = psg_analysed_data['analysed']['spo2_events']['data']
        spo2_flow_events['category'] = psg_analysed_data['analysed']['spo2_events']['category']

        # Represented in the data as a list of tuples with start/end time 
        time_stamps_flow = psg_analysed_data['analysed']['spo2_events']['timestamp']

        # Getting start times for the events 
        timestamps = [time_range[0] for time_range in time_stamps_flow]
        spo2_flow_events['timestamp'] = timestamps

        # Getting end times for the events 
        end_timestamps = [time_range[1] for time_range in time_stamps_flow]
        spo2_flow_events['timestamp_end'] = end_timestamps

        # Syncing timestamps
        spo2_flow_events['timestamp'] = self.sync_psg(spo2_flow_events['timestamp'])
        spo2_flow_events['timestamp_end'] = self.sync_psg(spo2_flow_events['timestamp_end'])

        return psg_analysed, psg_flow_events, spo2_flow_events

    def read_psg_derived(self):
        """
        Reads the psg_derived pickle.

        :return:
        :rtype: pd.DataFrame
        """
        psg_derived_file_name = "psg_derived_{}.pkl"
        psg_derived_data = pickle.load(open(self.path + '/psg/' + psg_derived_file_name.format(self.psg_id), 'rb'))
        psg_derived = pd.DataFrame()
        psg_derived['respiration_rate'] = psg_derived_data['derived']['respiration_rate_from_rip']['respiration_rate']
        psg_derived['timestamp'] = psg_derived_data['derived']['respiration_rate_from_rip']['timestamp']
        psg_derived['timestamp'] = self.sync_psg(psg_derived['timestamp'])
        return psg_derived

    def read_psg_raw(self):
        """
        Reads the psg_derived pickle.

        :return:
        :rtype: pd.DataFrame
        """
        psg_raw_file_name = "psg_raw_{}.pkl"
        psg_raw_data = pickle.load(open(self.path + '/psg/' + psg_raw_file_name.format(self.psg_id), 'rb'))
        psg_raw = pd.DataFrame()
        psg_raw['rip_thorax'] = psg_raw_data['raw']['rip_thorax']['data']
        psg_raw['rip_abdomen'] = psg_raw_data['raw']['rip_abdomen']['data']

        psg_raw['timestamp'] = psg_raw_data['raw']['rip_thorax']['timestamp']
        psg_raw['timestamp'] = self.sync_psg(psg_raw['timestamp'])

        return psg_raw

    def read_spo2(self):

        psg_raw_file_name = "psg_raw_{}.pkl"
        psg_raw_data = pickle.load(open(self.path + '/psg/' + psg_raw_file_name.format(self.psg_id), 'rb'))

        spo2 = pd.DataFrame()
        spo2['data'] = psg_raw_data['raw']['spo2']['data']
        spo2['timestamp'] = psg_raw_data['raw']['spo2']['timestamp']
        return spo2

    def get_first_and_last_timestamps(self):
        """
        Looks through all loaded datasets and find latest first and earliest last timestamps.
        Note the first and last timestamp of the actual PSG hypnogram is not included.

        :return: Two timestamps.
        :rtype: (pd.Timestamp, pd.Timestamp)
        """
        first_timestamps = []
        last_timestamps = []

        for df_name in self.loaded_data_frames:

            if df_name in ["psg_flow_events", "spo2_flow_events"]:
                pass
            else:
                first_ts = getattr(self, df_name).iloc[0, getattr(self, df_name).columns.get_loc('timestamp')]
                last_ts = getattr(self, df_name).iloc[-1, getattr(self, df_name).columns.get_loc('timestamp')]
                first_timestamps.append(first_ts)
                last_timestamps.append(last_ts)

        first = max(first_timestamps)
        last = min(last_timestamps)
        return first, last

    def clip_data(self):
        """
        Iterates through all datasets and clips them so they are all contained within the same two timestamps.

        :return: None
        """
        first, last = self.get_first_and_last_timestamps()

        # Not cutting wake periods at beginning and end
        first = self.psg.loc[self.psg['timestamp'] >= first, 'timestamp'].iloc[0]
        last = self.psg.loc[self.psg['timestamp'] <= last, 'timestamp'].iloc[-1]

        self.psg = self.psg.loc[self.psg['timestamp'] <= last, :]
        self.psg = self.psg.loc[self.psg['timestamp'] >= first, :]

        last = last + pd.Timedelta('30S')

        for df_name in self.loaded_data_frames:
            if df_name not in ["psg_flow_events", "spo2_flow_events"]:
                setattr(self, df_name, getattr(self, df_name).loc[getattr(self, df_name)['timestamp'] <= last, :])
                setattr(self, df_name, getattr(self, df_name).loc[getattr(self, df_name)['timestamp'] >= first, :])

    def set_epoch_indexes(self):
        """
        Adds a column to all datasets. The column sets for each data point what interval in the PSG hypnogram timestamps
        the data points falls within.

        :return: None
        """
        for df_name in self.loaded_data_frames:

            if df_name not in ["psg_flow_events", "spo2_flow_events"]:

                timestamp_values = self.psg['timestamp'].values

                attr_1 = getattr(self, df_name)
                if not attr_1.empty:
                    attr_2 = getattr(attr_1, 'timestamp')
                    attr_3 = getattr(attr_2, 'values')

                    epoch_index = self.get_epoch_index(
                        timestamp_values,
                        attr_3
                    )

                    setattr(self, df_name, getattr(self, df_name).assign(epoch_index=epoch_index))

    def sync_psg(self, timestamp):
        """
        Syncs the PSG to the Somnofy data so the timestamps are comparable. Only used to modify the PSG data timestamps.

        :param np.ndarray timestamp:
        :return: A np.ndarray with timestamps.
        :rtype: np.ndarray
        """
        if self.time_sync['offset_us'] > 0:
            timestamp = [
                i + datetime.timedelta(microseconds=abs(self.time_sync['offset_us'])) for
                i in timestamp]
        elif self.time_sync['offset_us'] < 0:
            timestamp = [
                i - datetime.timedelta(microseconds=abs(self.time_sync['offset_us'])) for
                i in timestamp]
        return timestamp

    @staticmethod
    def get_epoch_index(epoch_time, timestamp):
        """
        Returns a array that for each time point in timestamp specifies which interval of epoch_time it falls within.

        :param np.ndarray epoch_time: A np.ndarray with timestamps belonging to PSG hypnogram
        :param np.ndarray timestamp:  A np.ndarray with timestamps belonging to any dataset
        :return: A np.ndarray with timestamps with int.
        :rtype: np.ndarray
        """
        epoch_index = []
        for i in range(len(epoch_time) - 1):
            epoch_index += [i] * np.sum(
                np.logical_and(timestamp >= epoch_time[i], timestamp < epoch_time[i + 1]))
        epoch_index += [len(epoch_time) - 1] * np.sum(timestamp >= epoch_time[len(epoch_time) - 1])
        return epoch_index
