import pickle
import sys

import neurokit
import numpy as np
import os
import pandas as pd
import torch
from sklearn.preprocessing import MinMaxScaler, StandardScaler
from tqdm import tqdm

from Sensei.Session import Session


def get_data(path, somnofy_ids, psg_ids, categories, features, frequency=5):
    X = []
    Y = []

    progress_bar = tqdm(range(len(somnofy_ids)), desc="Sessions", file=sys.stdout)
    for i in progress_bar:

        # try:

        session, valid = load_session(path, somnofy_ids[i], psg_ids[i])

        if valid:
            data = extract_features(session, features)
            label = extract_label(session, data, categories)

            data.drop(['timestamp'], axis=1, inplace=True)

            print("data length: ", len(data), "and features: ", len(data[0]))

            X.append(data)
            Y.append(label)

        else:
            print("Dropped file with ids Somnofy. Not valid: ", somnofy_ids[i], "PSG: ", psg_ids[i])

        # except:
        #    print("Dropped file with ids Somnofy: Unknown ", somnofy_ids[i], "PSG: ", psg_ids[i])

    return X, Y


def extract_label(session, data, categories):
    labels = extract_flow_events_to_millisecond_granularity(data, session.psg_flow_events, categories)

    # Fill columns without a sickness
    labels['data'].fillna(0, inplace=True)

    labels['category'].fillna('Healthy', inplace=True)

    dummies = pd.get_dummies(labels['category'])
    dummies_categories = list(dummies.columns.values)
    missing_categories = [cat for cat in categories if cat not in dummies_categories]
    for cat in missing_categories:
        dummies[cat] = 0

    # Make sure column ordering is consistent
    ordered_categories = ['Healthy'] + categories
    dummies = dummies[ordered_categories]

    # KERAS specs
    labels_onehot = dummies.values.reshape(1, -1, dummies.shape[1])

    return labels_onehot


def extract_features(session, features, use_psg_raw=False, use_only_derived=False, frequency=5):
    extract_handcrafted_features = False
    extract_vitalsigns_features = False
    extract_derived_features = False
    extract_env_features = False
    extract_psg_features = False
    extract_psg_raw_features = False
    extract_spo2_features = False
    extract_spo2_flow_events_features = False

    if 'handcrafted' in features:
        extract_handcrafted_features = True

    if 'vitalsigns' in features:
        extract_vitalsigns_features = True

    if 'derived' in features:
        extract_derived_features = True
        df_der = session.derived

    if 'environment' in features:
        extract_env_features = True
        df_env = session.environment

    if 'psg' in features:
        extract_psg_features = True
        df_psg = session.psg_analysed

    if 'psg_raw' in features:
        extract_psg_raw_features = True
        df_psg_raw = session.psg_raw

    if 'spo2' in features:
        extract_spo2_features = True
        df_spo2 = session.spo2

    if 'spo2_flow_events' in features:
        extract_spo2_flow_events_features = True
        df_spo2_flow_events = session.spo2_flow_events

    if use_only_derived:
        df = session.derived.copy(deep=True)
        df['timestamp'] = df['timestamp'].astype('datetime64[ms]')

        df_psg_raw = session.psg_raw[['timestamp', 'rip_thorax']]
        df_psg_raw['timestamp'] = df['timestamp'].astype('datetime64[ms]')

        df = pd.merge(df, df_psg_raw, on='timestamp', how='left')

        return df

    df = session.vitalsigns.copy(deep=True)

    # VITALSIGNS
    # Convert timestamp in vitalsigns to granularity of milliseconds
    df['timestamp'] = df['timestamp'].astype('datetime64[ms]')
    df.reset_index(inplace=True)

    # Drop columns not in features. Timestamp needed for merge.
    if extract_vitalsigns_features:
        vt_features_to_keep = []
        vt_features_to_keep.extend(features['vitalsigns'])
        vt_features_to_keep.append('timestamp')
        df = df[vt_features_to_keep]

    # DERIVED RESPIRATION WAVEFORM
    if extract_derived_features:
        # Convert timestamps to milliseconds
        df_der['timestamp'] = df_der['timestamp'].astype('datetime64[ms]')

        # Choose which features to include
        der_features_to_keep = []
        der_features_to_keep.extend(features['derived'])
        der_features_to_keep.append('timestamp')
        df_der = df_der[der_features_to_keep]

        # Resample to frequency
        sampling_pace = int((1 / frequency) * 1000)
        df_der = df_der.resample('{}ms'.format(sampling_pace), on='timestamp').mean()

        # Merge derived with dataframe - NB: important to merge df into df_der, so that df is upsampled so df_der doesn't lose data
        df = pd.merge_asof(df_der, df, on='timestamp', direction='nearest', tolerance=pd.Timedelta('29ms'))
        # Interpolates missing values
        df.interpolate(method='linear', limit_direction='both', inplace=True)

    # ENVIRONMENT
    if extract_env_features:
        df_env['timestamp'] = df_env.timestamp.apply(lambda x: x.replace(microsecond=0))

        # Drop columns we don't need
        env_features_to_keep = []
        env_features_to_keep.extend(features['environment'])
        env_features_to_keep.append('timestamp')
        df_env = df_env[env_features_to_keep]

        df = pd.merge(df, df_env, on='timestamp', how='left')

        # Extrapolate the env features with 30sek granularity to 1 sek. NB !!!! PROBABLY NOT VALID ANYMORE !!!!
        df.interpolate(method='nearest', limit_direction='both', inplace=True)
        df.interpolate(method='linear', limit_direction='both', axis=0, limit=30, inplace=True)

    # HEART RATE
    if extract_psg_features:
        df_psg['timestamp'] = df_psg.timestamp.apply(lambda x: x.replace(microsecond=0))
        df_psg = df_psg[['heart_rate', 'timestamp']]

        df = pd.merge(df, df_psg, on='timestamp', how='left')

        # Drop duplicates due to merge
        df = df.drop_duplicates(subset=['timestamp'])

        # Interpolate missing HR values ( often due to movement )
        df['heart_rate'].interpolate(method='linear', limit_direction='both', axis=0, limit=20, inplace=True)

        print("extracted psg_features")

    if extract_psg_raw_features:
        # df_psg_raw['timestamp'] = df_psg_raw.timestamp.apply(lambda x: x.replace(microsecond=0))
        # df_psg_raw = df_psg_raw.drop_duplicates(subset=['timestamp'])

        psg_raw_features_to_keep = []
        psg_raw_features_to_keep.extend(features['psg_raw'])
        psg_raw_features_to_keep.append('timestamp')

        df_psg_raw = df_psg_raw[psg_raw_features_to_keep]

        if use_psg_raw:
            # ONLY TO BE USED DURING DEVELOPMENT
            return df_psg_raw

        df = pd.merge_asof(df, df_psg_raw, on='timestamp', direction='nearest')

    if extract_handcrafted_features:

        processed_respiration = None

        features_to_extract = features['handcrafted']

        if 'rsp_filtered' in features_to_extract:
            processed_respiration = neurokit.rsp_process(df['relative_distance'], sampling_rate=5)

            df['rsp_filtered'] = processed_respiration['df']['RSP_Filtered']

        if 'rsp_rate' in features_to_extract:
            if processed_respiration is None:
                processed_respiration = neurokit.rsp_process(df['relative_distance'], sampling_rate=5)

            df['rsp_rate'] = processed_respiration['df']['RSP_Rate']

        if 'rsp_broad' in features_to_extract:

            window_size = 600 * frequency  # seconds times frequency
            data = df['rsp_rate']
            signal = [np.nan] * int(np.round(window_size / 2))

            for index in range(0, len(data) - window_size):
                window = data[index: index + window_size]

                mean_rsp_rate = np.mean(window)

                signal.append(mean_rsp_rate)

            signal.extend([np.nan] * int(np.round(window_size / 2)))

            df['rsp_broad'] = signal

            df['rsp_broad'].interpolate(method='linear', limit_direction='both', inplace=True)
            df['rsp_rate'].interpolate(method='linear', limit_direction='both', inplace=True)

    if extract_spo2_features:
        # Convert timestamps to milliseconds
        df_spo2['timestamp'] = df_spo2['timestamp'].astype('datetime64[ms]')

        df_spo2 = df_spo2[['data', 'timestamp']]
        df_spo2.rename(columns={'data': 'spo2'}, inplace=True)

        # Resample to frequency
        sampling_pace = int((1 / frequency) * 1000)
        df_spo2 = df_spo2.resample('{}ms'.format(sampling_pace), on='timestamp').mean()

        # Merge derived with dataframe - NB: important to merge df into df_der, so that df is upsampled so df_der doesn't lose data
        df = pd.merge_asof(df_spo2, df, on='timestamp', direction='nearest')
        # Interpolates missing values
        df.interpolate(method='linear', limit_direction='both', inplace=True)

    if extract_spo2_flow_events_features:
        categories = ['Desaturation']
        if df_spo2_flow_events.empty:
            print("Spo2 dataframe empty, no events for session ", session.psg_id)
        else:
            df = extract_spo2_events_to_milliseconds(df, df_spo2_flow_events, granularity="ms")

    print(list(df))

    df.fillna(0, inplace=True)

    return df


def extract_spo2_events_to_milliseconds(df, dataframe, granularity="ms"):
    target_columns = ['data', 'timestamp']
    all_data = []

    for index, row in dataframe.iterrows():

        time_range = pd.date_range(start=row['timestamp'], end=row['timestamp_end'], freq=granularity)

        for timestamp in time_range:
            all_data.append([row['data'], timestamp])

    target_df = pd.DataFrame(all_data, columns=target_columns)
    target_df.rename(columns={'data': 'spo2_reduction'}, inplace=True)

    df = pd.merge_asof(df, target_df, on='timestamp', direction='nearest', tolerance=pd.Timedelta('29ms'))

    return df


def extract_flow_events_to_second_granularity(data, session, categories):
    return extract_flow_events_to_granularity(data, session, categories, granularity='S')


def extract_flow_events_to_millisecond_granularity(data, session, categories):
    return extract_flow_events_to_granularity(data, session, categories, granularity='ms')


def extract_flow_events_to_granularity(data, dataframe, categories, granularity="S", drop_timestamp=True):
    target_columns = ['data', 'category', 'timestamp']
    all_data = []

    for index, row in dataframe.iterrows():
        if row['category'] in categories:

            time_range = pd.date_range(start=row['timestamp'], end=row['timestamp_end'], freq=granularity)

            for timestamp in time_range:
                all_data.append([row['data'], row['category'], timestamp])

    target_df = pd.DataFrame(all_data, columns=target_columns)

    feature_target_merge = pd.merge(data, target_df, on='timestamp', how='left')

    labels = feature_target_merge[target_columns]

    if drop_timestamp:
        labels.drop(columns=['timestamp'], axis=1, inplace=True)

    return labels


#
# def scale_data(feature):
#     minMaxScaler = MinMaxScaler(feature_range=(0, 1))
#     try:
#         feature = minMaxScaler.fit_transform(feature)
#     except:
#         pass
#
#     return feature
#
#

def scale_data(X, mappings, scalers=None):
    if scalers == None:
        scalers = dict()

        night_indexes = []
        count = 0
        for night in X:
            count += len(night)
            night_indexes.append(count)
        x_flat = np.vstack(X)
        for index in mappings.keys():

            feature = mappings[index]

            if feature != 'spo2_events' or feature != 'signal_quality':
                scalers[feature] = StandardScaler(copy=False)
            else:
                scalers[feature] = MinMaxScaler(copy=False)

            column_data = x_flat[:, index].reshape(-1, 1)
            scaled_column_data = scalers[feature].fit_transform(column_data).reshape(-1)
            x_flat[:, index] = scaled_column_data

        res = [x_flat[i: j] for i, j in zip([0] +
                                            night_indexes, night_indexes + [None])]
        res.pop()
        return res, scalers

    else:

        night_indexes = []
        count = 0
        for night in X:
            count += len(night)
            night_indexes.append(count)
        x_flat = np.vstack(X)

        for index in range(0, len(X[0][0])):
            feature = mappings[index]

            column_data = x_flat[:, index].reshape(-1, 1)
            scaled_column_data = scalers[feature].transform(column_data).reshape(-1)
            x_flat[:, index] = scaled_column_data

        res = [x_flat[i: j] for i, j in zip([0] +
                                            night_indexes, night_indexes + [None])]
        res.pop()
        return res, scalers


def load_session(path, session_id, psg_id):
    """
    Creates a session object which loads all needed data specified for a session.

    :param str path: Path to processed folder.
    :param str session_id: Somnofy session ID e.g. VTNJQxICBxUJAwAA.
    :param str psg_id: PSG ID e.g. 7412.
    :return: A Session object.
    :rtype: hackfest.Session.Session
    """

    valid = True
    session = Session(path=path, session_id=session_id, psg_id=psg_id)
    session.psg.reset_index(inplace=True, drop=True)

    try:
        session.set_epoch_indexes()
    except LookupError:  # Something weird with PSG file 1987 that wouldn't read in properly.
        valid = False
        print("session ", session_id, " not valid..")

    return session, valid


def reshape_data_for_keras(feature):
    feature = feature.values.reshape(1, -1, feature.shape[1])
    return feature


def dump_data_to_file(filename, X, Y):
    # For dumping data
    data = [X, Y]

    with open(filename, 'wb') as outfile:
        pickle.dump(data, outfile)


def load_data_from_file(filename):
    with open(os.path.join(os.path.dirname(__file__), filename), 'rb') as infile:
        data = pickle.load(infile)

    return data[0], data[1]


def format_data_for_pytorch_cnn(X, Y, window_size, stride, number_of_features, is_test_data):
    data = []
    labels = []

    indexes = {
        0: [],
        1: [],
        2: []
    }

    added_segment_index = 0

    for i in tqdm(range(0, len(X))):

        if is_test_data:
            event_data = X[i]
        else:
            event_data = X[i][0]

        event_labels = Y[i][0]
        len_event_data = len(event_data)

        padded_event_data = event_data

        l = 0
        while l + window_size < len_event_data - 1:

            segment_data = padded_event_data[l: l + window_size]
            segment_labels = event_labels[l: l + window_size]

            if len(segment_labels) == window_size:
                corresponding_segment_label = np.where(segment_labels[int(window_size / 2)] == 1)[0][0]
            else:
                corresponding_segment_label = 0

            segment_data_torch = torch.from_numpy(segment_data).float().t().cuda()
            segment_labels_torch = torch.from_numpy(np.array([corresponding_segment_label])).long().cuda()

            if segment_data_torch.size()[1] == window_size:
                data.append(segment_data_torch)
            labels.append(segment_labels_torch)

            if not is_test_data:
                indexes[corresponding_segment_label].append(added_segment_index)
            added_segment_index += 1

            if is_test_data:
                l += 10  # IF YOU STEP WITH 1 THEN IT'S TOO MUCH DATA. MANY GIGS.
            else:
                l += 5
    if not is_test_data:
        return data, labels, indexes

    else:
        return data, labels
