import datetime
import pickle

import os
import pandas as pd
import random
import torch
from sklearn import model_selection
from sklearn.metrics import classification_report, cohen_kappa_score
from sklearn.utils import compute_sample_weight
from torch.optim.lr_scheduler import MultiStepLR
from torch.utils.data import DataLoader

from Sensei.casemanager import build_case_manager, build_dataset_from_case_manager, add_healthy_cases
from Sensei.data import format_data_for_pytorch_cnn, scale_data
from Sensei.dataset_pytorch import CNNDataset
from Sensei.helpers import print_event_counts, load_bergen_and_colosseum_combined, \
    reduce_data_for_training, print_class_count, keep_only_select_columns
from Sensei.models_pytorch import SimpleCNN
from Sensei.present_results import create_confusion_matrix, plot_confusion_matrix
from Sensei.sampler import ImbalancedSampler
from Sensei.train_torch import train_net, run_post_training_analysis

random.seed(3)

print("IS CUDA AVAILABLE:", torch.cuda.is_available())
torch.cuda.set_device(1)

path_bergen = "processed_data_files_bergen/bergen_all_features"
path_colosseum = "processed_data_files_colosseum/colosseum_all_features"

categories = ['Healthy', 'Hypopnea', 'OSA']

nr_features = 12

window_size_by_seconds = 30  # seconds
stride_by_seconds = 1  # seconds
frequency = 5  # Hz
window_size = window_size_by_seconds * frequency  # seconds * frequency
stride = stride_by_seconds * frequency  # seconds * frequency
pre_event_buffer = int(window_size / 2) + 1
post_event_buffer = int(window_size / 2) + 1

n_epochs = 300
class_sample_weighting = [1, 1, 1]
test_every_n_epo = 4
save_every_n_epo = 4

random.seed(4)
torch.manual_seed(4)

VERBOSE = True
REDUCE_DATA_FOR_TRAINING = True

sampling_technique = "oversampling"

# pretrained_cnn_model_path = "model_saves_cnn/1909-17052019/model_final"  # WITH new feature set ;) ;)

mappings = {0: 'spo2',
            1: 'relative_distance',
            2: 'sensor_state',
            3: 'movement_fast',
            4: 'movement_power_fast',
            5: 'movement_slow',
            6: 'movement_power_slow',
            7: 'signal_quality',
            8: 'distance',
            9: 'respiration_rate',
            10: 'respiration_distance',
            11: 'rsp_filtered',
            12: 'rsp_rate',
            13: 'rsp_broad',
            14: 'spo2_reduction'}

columns_to_keep = [0, 2, 3, 5, 6, 7, 8, 9, 10, 12, 13, 14]

kfold = model_selection.KFold(n_splits=5, shuffle=True, random_state=3)

results_df = pd.DataFrame()

X, Y = load_bergen_and_colosseum_combined(path_bergen, path_colosseum, shuffle_seed=3)

data_all_runs = []

for fold_number, (train_index, test_index) in enumerate(kfold.split(X, Y)):
    print("***************** FOLD NUMBER {}****************".format(fold_number))

    X_train, Y_train = [X[index] for index in train_index], [Y[index] for index in train_index]
    X_test, Y_test = [X[index] for index in test_index], [Y[index] for index in test_index]

    X_train, scalers_train = scale_data(X_train, mappings)
    X_test, _ = scale_data(X_test, mappings, scalers_train)

    # X_train, Y_train = np.take(X, train_index), np.take(Y, train_index)
    # X_test, Y_test = np.take(X, test_index, axis=0), np.take(Y, test_index, axis=0)

    date = datetime.datetime.now().strftime("%H%M-%d%m%Y")
    save_path = "model_saves_cnn/{}".format(date)

    if REDUCE_DATA_FOR_TRAINING:
        X_train, Y_train, X_test, Y_test = reduce_data_for_training(X_train, Y_train, X_test, Y_test, size_train=10,
                                                                    size_test=4)
    if VERBOSE:
        print("For training data:")
        print_event_counts(Y_train, frequency)
        print("For test data:")
        print_event_counts(Y_test, frequency)
        print()

    X_train = keep_only_select_columns(X_train, columns_to_keep)
    X_test = keep_only_select_columns(X_test, columns_to_keep)

    print("Building casemanager for training data... \t", end="")
    case_manager = build_case_manager(X_train, Y_train, pre_event_buffer, post_event_buffer)
    case_manager = add_healthy_cases(X_train, Y_train, case_manager, window_size=30, min_length=5 * frequency,
                                     drop_healthy_cases_percentage=0)

    print("done!")

    print("Building training data... \t\t\t\t\t", end="")
    X_train, Y_train = build_dataset_from_case_manager(case_manager)
    print("done!")

    print("Formating data for pytorch... \t\t\t\t", end="")
    X_train, Y_train, indexes = format_data_for_pytorch_cnn(X_train, Y_train, window_size, stride, nr_features,
                                                            is_test_data=False)
    X_test, Y_test = format_data_for_pytorch_cnn(X_test, Y_test, window_size, stride, nr_features, is_test_data=True)
    print("done!")

    if VERBOSE: print_class_count(Y_train)

    print("Creating Datasets and dataloaders... \t\t", end="")
    dataset_train = CNNDataset(X_train, Y_train)
    dataset_test = CNNDataset(X_test, Y_test)
    imbalanced_sampler = ImbalancedSampler(dataset_train, indexes, class_sample_weighting,
                                           sampling_technique=sampling_technique)

    train_loader = DataLoader(dataset_train, sampler=imbalanced_sampler, batch_size=128, shuffle=False, num_workers=0)
    test_loader = DataLoader(dataset_test, batch_size=512, shuffle=False, num_workers=0)
    print("done!")

    cnn_model = SimpleCNN(nr_features, dropout=0.2).cuda()
    optimizer = torch.optim.Adam(cnn_model.parameters(), lr=0.005)
    scheduler = MultiStepLR(optimizer, milestones=[30, 80, 130], gamma=0.1)

    loss = torch.nn.CrossEntropyLoss()

    net, stats = train_net(cnn_model, train_loader, test_loader, n_epochs, optimizer, loss, test_every_n_epo,
                           save_every_n_epo, save_path, scheduler, verbose=VERBOSE, early_stopping=True,
                           early_stopping_patience=20)

    print("Saving model... ", end="")
    torch.save(net, os.path.join(save_path, "model_final"))
    torch.save(net, "model_saves_cnn/cnn_latest")
    print("done!")

    print("Starting post-training analysis... ")

    best_epoch_save_path = os.path.join(save_path, "model_latest_epo_{}".format(stats['best_epo']))
    print("Using epo save: ", best_epoch_save_path)
    net = SimpleCNN(nr_features).cuda()
    old_model = torch.load(best_epoch_save_path)
    old_model_dict = old_model.state_dict()
    net.load_state_dict(old_model_dict)

    data = run_post_training_analysis(net, test_loader, categories, stats=stats, stride_by_seconds=stride_by_seconds,
                                      verbose=VERBOSE, plot_graphs=True)

    data_all_runs.append(data)
    print("... done!")

    print("Done at ", datetime.datetime.now(), " undersampling")

    del train_loader
    del test_loader
    del case_manager
    del imbalanced_sampler
    del dataset_train
    del optimizer
    del cnn_model
    del dataset_test
    del loss
    del net
    del stats
    del X_train
    del old_model
    del old_model_dict
    del X_test
    del Y_train
    del Y_test
    del scheduler
    torch.cuda.empty_cache()

print("SUMMARY K-FOLD")
print("#" * 70)

all_preds = []
all_labels = []

predicted_events_list = []
labeled_events_list = []
predicted_AHI_list = []
labeled_AHI_list = []

for run_data in data_all_runs:
    all_preds.extend(run_data['predictions'])
    all_labels.extend(run_data['labels'])

    predicted_events_list.append(run_data["events"]["predicted_events"])
    labeled_events_list.append(run_data["events"]["labeled_events"])
    predicted_AHI_list.append(run_data["ahi"]["predicted_AHI"])
    labeled_AHI_list.append(run_data["ahi"]["labeled_AHI"])

sample_weights = compute_sample_weight("balanced",
                                       all_preds)

rapport = classification_report(all_labels, all_preds)  # , sample_weight=sample_weights)
conf_mat = create_confusion_matrix(all_labels, all_preds)
cohens_cappa = cohen_kappa_score(all_labels, all_preds)  # , sample_weight=sample_weights)

print("\n Summary all tests... \t\t\t\t")
print("#" * 70, "\n")

print("Sklearn rapport\n", rapport)
print("Cohens kappa\n", cohens_cappa)
print("\nConfusion matrix\n", conf_mat)

plot_confusion_matrix(conf_mat, categories)

date = datetime.datetime.now().strftime("%H%M-%d%m%Y")
save_path = "cnn_k_fold_saves/{}".format(date)
print("\nSaving data from tests in ", save_path)
with open(save_path, 'wb') as outfile:
    pickle.dump(data_all_runs, outfile)
