import datetime
import pickle

import numpy as np
import os
import random
import torch
from sklearn.metrics import classification_report, cohen_kappa_score
from sklearn.utils import compute_sample_weight
from torch.optim.lr_scheduler import MultiStepLR
from torch.utils.data import DataLoader

from Sensei.casemanager import build_case_manager, build_dataset_from_case_manager, add_healthy_cases
from Sensei.data import format_data_for_pytorch_cnn, scale_data
from Sensei.dataset_pytorch import CNNDataset
from Sensei.helpers import print_event_counts, load_bergen_and_colosseum_combined, \
    reduce_data_for_training, print_class_count, split_train_test, keep_only_select_columns
from Sensei.models_pytorch import SimpleCNN
from Sensei.present_results import create_confusion_matrix, plot_confusion_matrix
from Sensei.sampler import ImbalancedSampler
from Sensei.train_torch import train_net, run_post_training_analysis

random.seed(3)

print("IS CUDA AVAILABLE:", torch.cuda.is_available())
torch.cuda.set_device(0)

path_bergen = "processed_data_files_bergen/bergen_all_features"
path_colosseum = "processed_data_files_colosseum/colosseum_all_features"

categories = ['Healthy', 'Hypopnea', 'OSA']

nr_features = 12  # 15
train_test_split = 0.15
validation_split = 0.15

window_size_by_seconds = 30  # seconds
stride_by_seconds = 1  # seconds
frequency = 5  # Hz
window_size = window_size_by_seconds * frequency  # seconds * frequency
stride = stride_by_seconds * frequency  # seconds * frequency
pre_event_buffer = int(window_size / 2) + 1
post_event_buffer = int(window_size / 2) + 1

n_epochs = 100
class_sample_weighting = [1, 1, 1]
test_every_n_epo = 4
save_every_n_epo = None

sampling_technique = None

random.seed(4)
torch.manual_seed(4)

VERBOSE = True

should_train = True
should_test = False
should_print_graphs_for_every_night = False

REDUCE_DATA_FOR_TRAINING = False

pretrained_cnn_model_path = "model_saves_cnn/0924-21052019/model_latest_epo_73"  # WITH new feature set ;) ;)

mappings = {0: 'spo2',
            1: 'relative_distance',
            2: 'sensor_state',
            3: 'movement_fast',
            4: 'movement_power_fast',
            5: 'movement_slow',
            6: 'movement_power_slow',
            7: 'signal_quality',
            8: 'distance',
            9: 'respiration_rate',
            10: 'respiration_distance',
            11: 'rsp_filtered',
            12: 'rsp_rate',
            13: 'rsp_broad',
            14: 'spo2_reduction'}

columns_to_keep = [0, 2, 3, 5, 6, 7, 8, 9, 10, 12, 13,
                   14]  # After what looks good with randoforrest

if should_train:
    date = datetime.datetime.now().strftime("%H%M-%d%m%Y")
    save_path = "model_saves_cnn/{}".format(date)

    X, Y = load_bergen_and_colosseum_combined(path_bergen, path_colosseum, shuffle_seed=5)

    X_r, X_validation, Y_r, Y_validation = split_train_test(X, Y, validation_split)

    X_train, X_test, Y_train, Y_test = split_train_test(X_r, Y_r, train_test_split)

    X_train, scalers_train = scale_data(X_train, mappings)
    X_test, scalers_test = scale_data(X_test, mappings, scalers_train)

    save_path_scalers = "cnn_training_scalers/{}".format(date)
    print("\nSaving scaler in ", save_path_scalers)
    with open(save_path_scalers, 'wb') as outfile:
        pickle.dump(scalers_train, outfile)

    if REDUCE_DATA_FOR_TRAINING:
        X_train, Y_train, X_test, Y_test = reduce_data_for_training(X_train, Y_train, X_test, Y_test, size_train=3,
                                                                    size_test=2)

    if VERBOSE:
        print("For training data:")
        print_event_counts(Y_train, frequency)
        print("For test data:")
        print_event_counts(Y_test, frequency)
        print()
        print("For validation data:")
        print_event_counts(Y_validation, frequency)
        print()

    X_train = keep_only_select_columns(X_train, columns_to_keep)
    X_test = keep_only_select_columns(X_test, columns_to_keep)

    print("Building casemanager for training data... \t", end="")
    case_manager = build_case_manager(X_train, Y_train, pre_event_buffer, post_event_buffer)
    case_manager = add_healthy_cases(X_train, Y_train, case_manager, window_size=30,
                                     min_length=5 * frequency,
                                     drop_healthy_cases_percentage=0)

    # case_manager = undersampling_random(case_manager)
    print("done!")

    print("Building training data... \t\t\t\t\t", end="")
    X_train, Y_train = build_dataset_from_case_manager(case_manager)
    print("done!")

    print("Formating data for pytorch... \t\t\t\t", end="")
    X_train, Y_train, indexes = format_data_for_pytorch_cnn(X_train, Y_train, window_size, stride, nr_features,
                                                            is_test_data=False)
    X_test, Y_test = format_data_for_pytorch_cnn(X_test, Y_test, window_size, stride, nr_features, is_test_data=True)
    print("done!")

    if VERBOSE: print_class_count(Y_train)

    print("Creating Datasets and dataloaders... \t\t", end="")
    dataset_train = CNNDataset(X_train, Y_train)
    dataset_test = CNNDataset(X_test, Y_test)
    imbalanced_sampler = ImbalancedSampler(dataset_train, indexes, class_sample_weighting,
                                           sampling_technique=sampling_technique)

    train_loader = DataLoader(dataset_train, sampler=imbalanced_sampler, batch_size=256, shuffle=False, num_workers=0)
    test_loader = DataLoader(dataset_test, batch_size=512, shuffle=False, num_workers=0)
    print("done!")

    cnn_model = SimpleCNN(nr_features, dropout=0.3, channels=15).cuda()
    optimizer = torch.optim.Adam(cnn_model.parameters(), lr=0.1)
    scheduler = MultiStepLR(optimizer, milestones=[20, 35, 50], gamma=0.1)

    loss = torch.nn.CrossEntropyLoss()

    net, stats = train_net(cnn_model, train_loader, test_loader, n_epochs, optimizer, loss, test_every_n_epo,
                           save_every_n_epo, save_path, scheduler, verbose=VERBOSE)

    print("Saving model... ", end="")
    torch.save(net, os.path.join(save_path, "model_final"))
    torch.save(net, "model_saves_cnn/cnn_latest")
    print("done!")

    print("Starting post-training analysis... ")

    run_post_training_analysis(net, stats, test_loader, categories, stride_by_seconds=stride_by_seconds,
                               verbose=VERBOSE, plot_graphs=True)
    print("... done!")

    print("LOOK AT WHEN TO STOP THE MODEL!!!! USE THIS nr of epochs in pipeline validation_run")

#
#
#
#

if should_test:

    print("Testing on Colosseum-data...")

    net = SimpleCNN(nr_features).cuda()
    old_model = torch.load(pretrained_cnn_model_path)
    old_model_dict = old_model.state_dict()
    net.load_state_dict(old_model_dict)

    X, Y = load_bergen_and_colosseum_combined(path_bergen, path_colosseum, shuffle_seed=1)

    _, X_test, _, Y_test = split_train_test(X, Y, train_test_split)

    save_path_scalers = "cnn_training_scalers/0924-21052019"

    with open(os.path.join(os.path.dirname(__file__), save_path_scalers), 'rb') as infile:
        scalers_train = pickle.load(infile)

    X_test, scalers_test = scale_data(X_test, mappings, scalers_train)

    X_test = keep_only_select_columns(X_test, columns_to_keep)

    print("Number of training files: ", len(X_test))
    print("Data loaded from file colosseum and Bergen")

    predicted_events_list = []
    labeled_events_list = []
    predicted_AHI_list = []
    labeled_AHI_list = []

    overall_cm = np.array([0] * 9).reshape((3, 3))

    data_all_runs = []

    for night_data, night_labels in zip(X_test, Y_test):
        night_data = [night_data]
        night_labels = [night_labels]

        print("Formating data for pytorch... \t\t\t\t", end="")
        X_test, Y_test = format_data_for_pytorch_cnn(night_data, night_labels, window_size, stride, nr_features,
                                                     is_test_data=True)
        print("done!")

        print("\tCreating dataset and dataloaders... ", end="")
        dataset = CNNDataset(X_test, Y_test)
        data_loader = DataLoader(dataset, batch_size=512, shuffle=False, num_workers=0)
        print("done!")

        print("\tRunning post training analysis... ", end="")

        data = run_post_training_analysis(net, data_loader,
                                          categories,
                                          stats=None,
                                          threshold=10 * frequency,
                                          stride_by_seconds=stride_by_seconds,
                                          verbose=VERBOSE,
                                          plot_graphs=should_print_graphs_for_every_night)
        data_all_runs.append(data)

    all_preds = []
    all_labels = []

    for run_data in data_all_runs:
        all_preds.extend(run_data['predictions'])
        all_labels.extend(run_data['labels'])

        predicted_events_list.append(run_data["events"]["predicted_events"])
        labeled_events_list.append(run_data["events"]["labeled_events"])
        predicted_AHI_list.append(run_data["ahi"]["predicted_AHI"])
        labeled_AHI_list.append(run_data["ahi"]["labeled_AHI"])

    sample_weights = compute_sample_weight("balanced",
                                           all_preds)

    rapport = classification_report(all_labels, all_preds, sample_weight=sample_weights)
    conf_mat = create_confusion_matrix(all_labels, all_preds)
    cohens_cappa = cohen_kappa_score(all_labels, all_preds, sample_weight=sample_weights)

    print("\n Summary all tests... \t\t\t\t")
    print("#" * 70, "\n")

    print("Sklearn rapport\n", rapport)
    print("Cohens kappa\n", cohens_cappa)
    print("\nConfusion matrix\n", conf_mat)

    plot_confusion_matrix(conf_mat, categories)

    # plot_testing_ahi_and_event_stats(predicted_events_list, predicted_AHI_list, labeled_events_list,
    #                                 labeled_AHI_list)

    date = datetime.datetime.now().strftime("%H%M-%d%m%Y")
    save_path = "cnn_pipeline_data_saves/{}".format(date)
    print("\nSaving data from tests in ", save_path)
    with open(save_path, 'wb') as outfile:
        pickle.dump(data_all_runs, outfile)
