import random

import torch
from torch.utils.data import Sampler


class ImbalancedSampler(Sampler):
    def __init__(self, dataset, indexes, class_sample_weighting=None, samples_per_class=30, sampling_technique=None,
                 random_seed=3):

        random.seed(3)

        self.dataset = dataset
        self.indexes = indexes
        self.class_sample_weighting = class_sample_weighting
        self.samples_per_class = samples_per_class

        if self.class_sample_weighting is None:
            self.class_sample_weighting = [1] * len(self.indexes)

        if sampling_technique is None:
            self.sampling_technique = self.__default_sampling__

            class_sizes = [len(class_indexes) for class_indexes in self.indexes.values()]
            class_sizes_without_healthy = class_sizes[1:]
            self.samples_per_class = max(class_sizes_without_healthy) * 2

        elif sampling_technique == 'oversampling':
            self.sampling_technique = self.__oversampling__

            class_sizes = [len(class_indexes) for class_indexes in self.indexes.values()]
            self.samples_per_class = max(class_sizes)

        elif sampling_technique == 'undersampling':
            self.sampling_technique = self.__undersampling__

            class_sizes = [len(class_indexes) for class_indexes in self.indexes.values()]
            self.samples_per_class = min(class_sizes)
            print("samples per class: ", self.samples_per_class)

        else:
            raise KeyError("{} did not map to any of the available sampling techniques".format(sampling_technique))

    def __iter__(self):
        return self.sampling_technique()

    def __len__(self):
        return len(self.dataset)

    def __default_sampling__(self):
        samples = []

        for values, class_sample_weight in zip(self.indexes.values(), self.class_sample_weighting):
            if values:
                # sample = random.samples(values, k=smallest_number_of_events) Without replacement

                weighted_count = int(self.samples_per_class * class_sample_weight) - 1

                sample = random.choices(values, k=weighted_count)  # With replacement
                samples.extend(torch.LongTensor(sample).cuda())

        random.shuffle(samples)

        return iter(samples)

    def __oversampling__(self):
        samples = []

        for indexes in self.indexes.values():
            if indexes:
                class_samples = random.choices(indexes, k=self.samples_per_class)

                samples.extend(class_samples)

        random.shuffle(samples)

        return iter(samples)

    def __undersampling__(self):
        samples = []

        for indexes, class_sample_weight in zip(self.indexes.values(), self.class_sample_weighting):
            if indexes:
                weighted_count = int(self.samples_per_class * class_sample_weight)

                class_samples = random.choices(indexes, k=weighted_count)

                samples.extend(class_samples)

        random.shuffle(samples)

        return iter(samples)
